def getPlaybackContext(splay=True, autoCaptionsDefaultOn=False, signatureTimestamp=19445, vid="hfyLjRZmEFc"):

    context = {
        'contentPlaybackContext': {
            'currentUrl': f'/watch?v={vid}',
            'vis': 0,
            'splay': splay,
            'autoCaptionsDefaultOn': False,
            'autonavState': 'STATE_OFF',
            'html5Preference': 'HTML5_PREF_WANTS',
            'signatureTimestamp': signatureTimestamp,
            'referer': f'https://www.youtube.com/watch?v={vid}',
            'lactMilliseconds': '1',
            'watchAmbientModeContext': {
                'watchAmbientModeEnabled': True
            }
        }
    }
    return context



import requests


from .context import Context

from bs4 import BeautifulSoup
class Request:
    def __init__(self):
        self.__Context = Context()
        self.__initHeaders()

    def __initHeaders(self):
        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Credentials": "include"
        }

    def addHeaders(self, headers):
        self.headers.update(headers)

    def removeHeaders(self, headerNames):
        for key in headerNames:
            self.headers.pop(key)

    def updateContextState(self, videoId=None, replacables={}, additional={}, removables=[]):
        if videoId is not None:
            self.__Context.replaceVideoId(videoId)
        for key, val in replacables.items():
            if key in self.__Context:
                self.__Context[key] = val
            else:
                self.__Context.replace(key, val)

        for key, val in additional.items():
            self.__Context.add(key, val)

        for key in removables:
            if key in self.__Context.get():
                self.__Context.remove(key)
            else:
                self.__Context.replace(key, '')

    def send(self, url, method="GET", additional_payload=None):

        if type(method) != str or method not in ["GET", "POST"]:
            raise ValueError(
                "Invalid 'method' value. must be either 'GET' or 'POST'")

        # if type(payload) != dict:
        #     raise ValueError("Invalid 'payload' value. must be a dictionary")

        if type(url) != str:
            raise ValueError("Invalid 'url' value. must be a string")

        # Merge payloads
        payload = self.__Context.get()
        # if additional_payload is not None and type(additional_payload) == dict:
        #     payt
        #     payload.update(additional_payload)
        if method.upper() == "GET":
            response = requests.get(url, headers = {
                "Accept":"text/html"
            })

        elif method.upper() == "POST":
            response = requests.post(
                url,
                json=payload,
                headers=self.headers
            )

        if response.status_code == 200:
            if method.upper() == "GET":
                return response
            elif method.upper() == "POST":
                return response.json()
        else:
            print(response.content)
            raise RuntimeError(f"Request failed ({response.status_code})")
        return None





import re
import json
import requests
from bs4 import BeautifulSoup

from .utils.handler import BaseHandler
from .utils.decorator import VideosScraper, CommentScraper, TranscriptScraper
from .utils.request import Request
from .utils.context import Context

# This package was built on hurry, I'll comment it later

class YTSpider:

    def __init__(self):
        self.Request = Request()
        self.__Response = {}
        self.__Context = Context()
        self.handlers = {}
        self.defaultHandlers = {}

    def attachHandler(self, handler):
        if not isinstance(handler, BaseHandler):
            raise ValueError(
                "Argument (handler) must be a subclass of BaseHandler")
        if not handler.name in self.defaultHandlers:
            self.defaultHandlers[handler.name] = handler

        self.handlers[handler.name] = handler

    def restoreDefaultHandlers(self, exclude=[]):
        for handlerName in self.defaultHandlers:
            if handlerName not in exclude:
                self.handlers[handlerName] = self.defaultHandlers[handlerName]

    def removeHandler(self, handler):
        if handler.name in self.handlers:
            del self.handlers[handler.name]
            return True
        return False

    def getHandler(self, handlerName):
        if handlerName in self.handlers:
            return self.handlers[handlerKey]
        return None

    def getAllHandlers(self):
        return self.handlers

    def executeHandlers(self, itemId):
        results = {}
        for handlerName, handler in self.handlers.items():

            result = handler.execute(itemId)
            if result is not None and result != []:
                results[handlerName] = result
        return results

    def scrape(self, itemId, continuation_token=None):
        self.__Response.clear()
        if type(itemId) == str:
            # self.itemId = itemId
            # self.__Response.update({itemId: self.scrapeSingle(itemId)})
            result = self.scrapeSingle(itemId)
            key = self.itemId if hasattr(self, "itemId") else itemId
            self.__Response.update({key: result})
        elif type(itemId) == list:
            for i, id in enumerate(itemId):
                if type(id) == str:
                    # self.itemId = id
                    # self.__Response.update({id: self.scrapeSingle(id)})
                    result = self.scrapeSingle(id)
                    key = self.itemId if hasattr(self, "itemId") else id
                    self.__Response.update({key: result})
                else:
                    raise ValueError(
                        f"Invalid data type at index {i}. Expected str")
        return self

    def scrapeSingle(self, itemId):
        result = {itemId: None}
        return result

    def get(self):
        for itemId in self.__Response.keys():
            self.itemId = itemId
            result = self.executeHandlers(itemId)

            if itemId in self.__Response:
                self.__Response[itemId].update(result)
            else:
                self.__Response[itemId] = result
        self.restoreDefaultHandlers()
        return self.__Response



class YTSVideo(YTSpider, CommentScraper, TranscriptScraper):
    def __init__(self, commentDefaultHandler=None, transcriptDefaultHandler=None):
        super(YTSVideo, self).__init__()
        CommentScraper.__init__(self, commentDefaultHandler)
        TranscriptScraper.__init__(self, transcriptDefaultHandler)
        self.url = "https://www.youtube.com/youtubei/v1/player?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false"

        #
    def scrapeSingle(self, itemId):
        self.Request.updateContextState(itemId)
        result = self.Request.send(self.url, "POST")
        if "videoDetails" in result:
            return result["videoDetails"]
        return None

class YTSChannel(YTSpider, VideosScraper):
    def __init__(self, videosDefaultHandler=None):
        super(YTSChannel, self).__init__()
        VideosScraper.__init__(self, videosDefaultHandler)
        self.url = "https://www.youtube.com/"

    def prepareURL(self, userInput):
        if type(userInput) != str:
            raise ValueError("Argument (screenName) must be a string")

        screenNames = re.findall(r"@\w+", userInput)
        if len(screenNames) > 0:
            screenName = screenNames[0]
        elif "youtube.com" in userInput:
            screenName = ""
            self.url = userInput
        else:
            raise ValueError(
                "Argument (screenName) must have a screen name as '@ChannelName' or a YouTube URL")
        return self.url + screenName.strip("/").strip()

    def __postProcessResponse(self, response):
        if isinstance(response, requests.models.Response):
            soup = BeautifulSoup(response.text)
            # print(list(map(lambda x: x['itemprop'], metas)))
            data = {}
            # FEATURE: add useful meta properties (attributes['property'])
            attributes = {
                "itemprop": [
                    "channelId",
                    "isFamilyFriendly",
                    "regionsAllowed",
                    "paid",
                    "name",
                    "url"
                ],
                "name": [
                    "description",
                    "keywords"
                ],
                "property": []
            }
            thumbnailUrl = soup.find("link", {"itemprop": "thumbnailUrl"})
            if thumbnailUrl is not None:
                thumbnailUrl = thumbnailUrl['href']

            for criteria in attributes:
                for metaKey in attributes[criteria]:
                    tag = soup.find("meta", {criteria: metaKey})
                    if tag is not None:
                        tagInnerContent = tag['content']
                        data[metaKey] = tagInnerContent
        
        # if "channelId" not in data:
        #     channelId = 
        # print(soup.find("meta", {"itemprop":"channelId"}))
            # channelIds = re.findall(r'"browseId":"([a-zA-Z0-9 _ -]+)"', str(soup))
            # if len(channelIds) > 0:
            #     data["channelId"] = channelIds[0]
        if "channelId" in data:
            self.itemId = data["channelId"]
            return data
        else:
            return None

    def scrapeSingle(self, screenName):
        endpoint = self.prepareURL(screenName)
        print("HHH")
        print(endpoint)
        result = self.Request.send(endpoint)
        data = self.__postProcessResponse(result)
        if not hasattr(self, "itemId"):
            self.itemId = screenName
        return data




from .handler import *


class ScrapingDecorator:
    def __init__(self, defaultHandler=BaseHandler()):

        if not isinstance(defaultHandler, BaseHandler):
            raise ValueError(
                "Argument (defaultHandler) must be a subclass of BaseHandler or 'scrape' or 'ignore'")
 
        self.defaultHandler = defaultHandler


    def setDefaultHandler(self, defaultHandler):
        if defaultHandler is None:
            raise ValueError(
                "Argument (defaultHandler) must be a subclass of parent (BaseHandler)")
        self.defaultHandlers[defaultHandler.name] = defaultHandler
        self.updateHandler(defaultHandler)

    def updateHandler(self, handler):

        if not isinstance(handler, BaseHandler):
            raise ValueError(
                "Argument (handler) must be a subclass of BaseHandler")
        self.attachHandler(handler)

class VideosScraper(ScrapingDecorator):
    def __init__(self, videosDefaultHandler=VideosDefaultHandler()):
        if videosDefaultHandler is None or commentDefaultHandler == 'ignore':
            videosDefaultHandler = VideosDefaultHandler()
        elif videosDefaultHandler == 'scrape':
            videosDefaultHandler = VideosScrapingHandler()

        super(VideosScraper, self).__init__(videosDefaultHandler)

        self.updateHandler(videosDefaultHandler)

    def withVideos(self, n_videos=29):

        handler = VideosScrapingHandler(n_videos=n_videos)
        self.updateHandler(handler)
        return self


class TranscriptScraper(ScrapingDecorator):
    def __init__(self, transcriptDefaultHandler=TranscriptionDefaultHandler()):
        if transcriptDefaultHandler is None or transcriptDefaultHandler == 'ignore':
            transcriptDefaultHandler = TranscriptionDefaultHandler()
        elif transcriptDefaultHandler == 'scrape':
            transcriptDefaultHandler = TranscriptionScrapingHandler()
        super(TranscriptScraper, self).__init__(transcriptDefaultHandler)
        self.updateHandler(transcriptDefaultHandler)

    def withTranscripts(self, only_in_langs=[]):

        handler = TranscriptionScrapingHandler(only_in_langs=only_in_langs)
        self.updateHandler(handler)
        return self

class CommentScraper(ScrapingDecorator):
    def __init__(self, commentDefaultHandler=CommentDefaultHandler()):
        if commentDefaultHandler is None or commentDefaultHandler == 'ignore':
            commentDefaultHandler = CommentDefaultHandler()
        elif commentDefaultHandler == 'scrape':
            commentDefaultHandler = CommentScrapingHandler()

        super(CommentScraper, self).__init__(commentDefaultHandler)

        self.updateHandler(commentDefaultHandler)

    def withComments(self, n_comments=15):

        handler = CommentScrapingHandler(n_comments=n_comments)
        self.updateHandler(handler)
        return self


from .request import Request

class BaseHandler:
    def __init__(self, name=None):
        self.request = Request()
        self.name = name

    def execute(self, itemId):

        return None



class CommentHandler(BaseHandler):
    def __init__(self, n_comments=0):
        super(CommentHandler, self).__init__(name='comments')
        self.n_comments = n_comments

    def execute(self, itemId):
        return None


class CommentDefaultHandler(CommentHandler):
    def __init__(self, n_comments=0):
        super(CommentDefaultHandler, self).__init__(n_comments)

    def execute(self, itemId):

        return None


class CommentScrapingHandler(CommentHandler):
    def __init__(self, n_comments=100):
        super(CommentScrapingHandler, self).__init__(n_comments)
        self.url = "https://www.youtube.com/youtubei/v1/next?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false"

    def execute(self, itemId):
        self.request.updateContextState(itemId, additional={"continuation": ""})
        firstTime = True
        token = ""
        all_comments = []
        while token is not None and len(all_comments) < self.n_comments:
            self.request.updateContextState(additional={"continuation": token})
            result = self.request.send(self.url, method="POST")
            if firstTime:
                # Try get token for comment retrieval
                try:
                    token = result["contents"]["twoColumnWatchNextResults"]["results"]["results"]["contents"][3]["itemSectionRenderer"]["contents"][0]["continuationItemRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
                    firstTime = False
                # Couldn't get token for comment retrieval
                except Exception as e:
                    token = None
            else:
                # Try get comments
                try:
                    comments = result["onResponseReceivedEndpoints"][-1]["reloadContinuationItemsCommand"]["continuationItems"][:-1]
                
                # There are no comments at all (not a single one!)
                except Exception as e:
                    comments = []
                
                # Try get token, but if comments are < 20, there is no 'continuation' token because there're no more comments
                try:
                    token = result["onResponseReceivedEndpoints"][-1]["reloadContinuationItemsCommand"]["continuationItems"][-1]["commentThreadRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
                
                # No more comments
                except Exception as e:
                    token = None

                for i in range(len(comments)):
                    comment_texts = comments[i]["commentThreadRenderer"]["comment"]["commentRenderer"]["contentText"]['runs']
                    comment = " ".join(
                        list(map(lambda c: c["text"], comment_texts)))
                    all_comments.append(comment)
        return all_comments



class TranscriptionHandler(BaseHandler):
    def __init__(self, only_in_langs=[]):
        super(TranscriptionHandler, self).__init__(name='transcripts')
        self.only_in_langs = only_in_langs

    def execute(self, itemId):
        return None


class TranscriptionScrapingHandler(TranscriptionHandler):
    def __init__(self, only_in_langs=[]):
        super(TranscriptionScrapingHandler, self).__init__(only_in_langs)

    def execute(self, itemId):
        result = ["One Two THree"]
        return result


class TranscriptionDefaultHandler(TranscriptionHandler):
    def __init__(self):
        super(TranscriptionDefaultHandler, self).__init__()

    def execute(self, itemId):
        return None


class VideosHandler(BaseHandler):
    def __init__(self, n_videos=0):
        super(VideosHandler, self).__init__(name='videos')
        self.n_videos = n_videos

    def execute(self, itemId):
        return None


class VideosDefaultHandler(VideosHandler):
    def __init__(self, n_videos=0):
        super(VideosDefaultHandler, self).__init__(n_videos)

    def execute(self, itemId):
        return None


class VideosScrapingHandler(VideosHandler):
    def __init__(self, n_videos=29):
        super(VideosScrapingHandler, self).__init__(n_videos)
        self.url = "https://www.youtube.com/youtubei/v1/browse?key=AIzaSyAO_FJ2SlqU8Q4STEHLGCilw_Y9_11qcW8&prettyPrint=false"

    def execute(self, itemId):
        params_token = ""
        firstTime = True
        token = ""
        all_videos = []

        self.request.updateContextState(additional={"browseId": itemId, "params": params_token})
        # Scrape MainPage tab
        result = self.request.send(self.url, "POST")

        try:
            # Get param which will route the scraping to Videos Tab
            tabs = result["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]
            if len(tabs) > 2 and tabs[1]["tabRenderer"]["title"] == "Videos":
                vidTabsEndPointParam = tabs[1]["tabRenderer"]["endpoint"]["browseEndpoint"]["params"]
                self.request.updateContextState(additional={"params": vidTabsEndPointParam})
                # Scrape Videos Tab
                result = self.request.send(self.url, "POST")

                tabs = result["contents"]["twoColumnBrowseResultsRenderer"]["tabs"]
                videoItems = tabs[1]["tabRenderer"]["content"]["richGridRenderer"]["contents"]
                # Remove unnecessary request payload
                self.request.updateContextState(removables=["params", "browseId"])
                while len(all_videos) < self.n_videos:
                    for videoItem in videoItems[:-1]:
                        vidRender = videoItem["richItemRenderer"]["content"]["videoRenderer"]
                        video = {}
                        video["id"] = vidRender["videoId"]
                        video["title"] = vidRender["title"]["runs"][0]["text"]
                        video["views_count"] = vidRender["viewCountText"]["simpleText"]
                        video["publish_at"] = vidRender["publishedTimeText"]["simpleText"]

                        if len(all_videos) >= self.n_videos:
                            return all_videos
                        all_videos.append(video)
                    try:
                        # Set Continuation Parameter
                        continuationParam = videoItems[-1]["commentThreadRenderer"]["continuationEndpoint"]["continuationCommand"]["token"]
                    except Exception as e:
                        print(e)
                        break
                    self.request.updateContextState(additional={"continuation":continuationParam})
                    result = self.request.send(self.url, "POST")
                    try:
                        videoItems = result["onResponseReceivedActions"][0]["appendContinuationItemsAction"]["continuationItems"]
                    except Exception as e:
                        print(e)
                        break

        except Exception as e:
            print(e)
            pass
        return all_videos

        # # TODO: Scrape for first time to get "params", then use tabs[1] to get params and scrape Videos tab[1]
        # # What about making one class 'ChannelTabHandler' which takes tab name (videos, playlists), nope because each has different output (solvable but code more for fun)




from setuptools import setup, find_packages


setup(
    name='ytspider',
    description="YouTube scraper, videos, channels, playlists, comments and transcriptions",
    version='0.0.0.3',
    license='MIT',
    author="Zeyad Khalid",
    maintainer_email='zeyad.khalid@must.edu.eg',
    author_email='zeyadpro99@gmail.com',
    package_dir={"ytspider":"src"},
    packages=find_packages(),#["ytspider","ytspider.utils"],
    py_modules=["ytspider.py"],
    url='https://github.com/zementalist',
    keywords='web crawling youtube scraping api',
    install_requires=[
          'requests',
          'bs4',
      ],
)


def getClient(lang="en", gl="EG", remoteHost="102.190.240.107", osName="Windows", osVersion="10.0", platform="Desktop"):
    client = {
        'hl': lang,
        'gl': gl,
        'remoteHost': remoteHost,
        'deviceMake': '',
        'deviceModel': '',
        'visitorData': 'Cgs5S3R6TGlJaHhYZyierrKhBg%3D%3D',
        'userAgent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/111.0.0.0 Safari/537.36,gzip(gfe)',
        'clientName': 'WEB',
        'clientVersion': '2.20230331.00.00',
        'osName': osName.upper(),
        'osVersion': osVersion,
        'originalUrl': 'https://www.youtube.com/watch?v=hfyLjRZmEFc',
        'platform': platform.upper(),
        'clientFormFactor': 'UNKNOWN_FORM_FACTOR',
        'configInfo': {'appInstallData': 'CJ6usqEGEOWg_hIQzJmvBRCCna8FEOf3rgUQzN-uBRDM9a4FEI73rgUQhYOvBRDks_4SELfgrgUQ4tSuBRDthq8FEL-QrwUQuNSuBRDMrv4SEKLsrgUQieiuBRC4i64FEIvqrgUQ1_-uBRC9tq4FEKC3_hIQ1Lj-Eg%3D%3D'},
        'timeZone': 'Africa/Cairo',
        'browserName': 'Chrome',
        'browserVersion': '111.0.0.0',
        'acceptHeader': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7',
        'deviceExperimentId': 'ChxOekl4T0RNeE1EUTFPRE13TlRNM01qUXpOZz09EJ6usqEGGJ6usqEG',
        'screenWidthPoints': 471,
        'screenHeightPoints': 592,
        'screenPixelDensity': 1,
        'screenDensityFloat': 1,
        'utcOffsetMinutes': 120,
        'userInterfaceTheme': 'USER_INTERFACE_THEME_LIGHT',
        'memoryTotalKbytes': '8000000',
        'clientScreen': 'WATCH',
        'mainAppWebInfo': {
            'graftUrl': '/watch?v=hfyLjRZmEFc',
            'pwaInstallabilityStatus': 'PWA_INSTALLABILITY_STATUS_UNKNOWN',
            'webDisplayMode': 'WEB_DISPLAY_MODE_BROWSER',
            'isWebNativeShareAvailable': True
        }
    }
    return client



import json
from .client import getClient
from .playback_context import getPlaybackContext
from .signals_info import getSignalsInfo

class Context:
    def __init__(self, contentCheckOk=False):
        self.contentCheckOk = contentCheckOk
        self.context = {
            'context':
            {
                'client': getClient("en"),
                'user': {'lockedSafetyMode': False},
                'request': {'useSsl': True,
                            'internalExperimentFlags': [],
                            'consistencyTokenJars': []},
                'clickTracking': {'clickTrackingParams': 'CDIQvU4YACITCLfOn-6Xkf4CFRPeFgodhAgHNTIJZW5kc2NyZWVuSOayt8W2i9D-9QGaAQUIAhD4HQ=='},
                'adSignalsInfo': getSignalsInfo(media="image"),
            },
            'videoId': 'hfyLjRZmEFc',
            'playbackContext': getPlaybackContext(splay=False),
            'racyCheckOk': False,
            'contentCheckOk': self.contentCheckOk}
        self.defaultId = "hfyLjRZmEFc"

    def replace(self, key, value):
        context_str = json.dumps(self.context)
        context_str = context_str.replace(key, value)
        self.context = json.loads(context_str)
        return self

    def replaceVideoId(self, value):
        self.replace(self.defaultId, value)
        self.defaultId = value
        return self

    def add(self, key, value):
        self.context[key] = value
        return self

    def remove(self, key):
        del self.context[key]
        return self

    def get(self):
        return self.context

    def getDefaultVideoId(self):
        return self.defaultId


